from math import copysign, log, isclose
from scipy.optimize import curve_fit, OptimizeWarning
import numpy as np
from warnings import warn, simplefilter

# Raise least square fitting warnings to exceptions
simplefilter('error', OptimizeWarning)
simplefilter('error', RuntimeWarning)

class ConvergenceCalculator:
    def __init__(self, h, phi):
        # Sort lists by grid size before assignment
        self.h, self.phi = (list(t) for t in zip(*sorted(zip(h, phi))))
        self.phi_extrapolated = None
        self.uncertainties = None
        self.fitted_function = None
        self.function_parameters = None
        self.execute()

    def __repr__(self):
        return 'ConvergenceCalculator({})'.format(
                ', '.join('{} = {}'.format(name, value)
                    for name, value in zip(['h', 'phi', 'phi_ext', 'U'],
                        [self.h, self.phi, self.phi_extrapolated, 
                            self.uncertainties
                            ]
                        )
                    )
                )

    def get_order(self):
        # Returns order of the fit. NaN is returned for mixed 1st/2nd order.
        if len(self.h) == 3:
            return self.calc_apparent_order()
        else:
            for f, p in zip([self.fitfunc_unknown_order,
                self.fitfunc_1st_order,
                self.fitfunc_2nd_order,
                self.fitfunc_1st_n_2nd_order], 
                [self.function_parameters[-1], 1, 2, float('nan')]
                ):
                if f == self.fitted_function:
                    return p

    def get_phi(self, h):
        # Returns value of the fitted function at h 
        if isinstance(h, int) and h == 0:
            return self.phi_extrapolated
        else:
            return self.fitted_function(h, *self.function_parameters)

    def execute(self):
        no_of_grids = len(self.h)
        if no_of_grids == 3:
            self.use_roache()
        elif no_of_grids > 3:
            self.use_eca_hoekstra()
        else:
            raise AttributeError('Insufficient number of solutions')

    def use_roache(self):
        self.phi_extrapolated = self.calc_phi_extrapolated()
        # Roache's method gives no uncertainty result for the coarsest mesh,
        # so an undefined placeholder is used to achieve expected code behaviour
        self.uncertainties = [*self.calc_gci(), float('nan')]
        self.fitted_function = self.fitfunc_roache_order
        self.function_parameters = self.calc_curve_fit()
        self.check_roache()

    def calc_apparent_order(self, error_limit = 1e-10):
        # Finds apparent order of convergence from the three finest grids
        # by application of Roache's method
        p_n = self.calc_p(0)
        q_n = self.calc_q(p_n)
        p_m = self.calc_p(q_n)
        while abs(p_n - p_m) > error_limit:
            p_n = p_m
            q_n = self.calc_q(p_n)
            p_m = self.calc_p(q_n)
        return p_m

    def calc_p(self, q):
        # Finds order of convergence from three finest grids for a given q
        differences = self.calc_differences()
        try:
            return next((1/log(r_i)) * abs(log(abs(e_j/e_i)) + q) 
                    for r_i, e_i, e_j in zip(
                        self.calc_refinement_factors()[:-1],
                        differences[:-1], 
                        differences[1:],
                        )
                    )
        except ZeroDivisionError:
            raise ValueError(' '.join(['Consecutive results are identical.',
                'Convergence can\'t be calculated.'
                ]))

    def calc_differences(self):
        return [phi_j - phi_i for phi_i, phi_j 
                in zip(self.phi[:-1], self.phi[1:])
                ]

    def calc_refinement_factors(self):
        return [h_j/h_i for h_i, h_j 
                in zip(self.h[:-1], self.h[1:])
                ]

    def calc_q(self, p):
        # Finds q from three finest grids for a given p
        refinement_factors = self.calc_refinement_factors()
        if p == 0.0:
            raise ValueError('Apparent order of convergence is zero')
        return next(log((r_i**p - s)/(r_j**p - s)) for r_i, r_j, s
                in zip(refinement_factors[:-1],
                    refinement_factors[1:],
                    self.calc_signs()
                    )
                )

    def calc_signs(self):
        differences = self.calc_differences()
        return [copysign(1.0, e_j/e_i) for e_i, e_j
            in zip(differences[:-1], differences[1:])
            ]

    def calc_phi_extrapolated(self):
        # Calculates extrapolated result from three finest grids
        p = self.calc_apparent_order()
        if p == 0.0:
            raise ValueError('Apparent order of convergence is zero.')
        return next(((r_i**p)*phi_i - phi_j)/(r_i**p - 1)
                for r_i, phi_i, phi_j in zip(
                    self.calc_refinement_factors(),
                    self.phi[:-1],
                    self.phi[1:]
                    )
                )

    def calc_gci(self, safety_factor = 1.25):
        p = self.calc_apparent_order()
        return [safety_factor*abs(e_a)/(r_i**p - 1) for e_a, r_i
                in zip(self.calc_approximate_errors(),
                    self.calc_refinement_factors(),
                    )
                ]

    def calc_approximate_errors(self):
        return [phi_i - phi_j for phi_i, phi_j
                in zip(self.phi[:-1], self.phi[1:])
                ]

    def fitfunc_roache_order(self, h, phi_0, alpha):
        return phi_0 + alpha*h**self.calc_apparent_order()

    def calc_curve_fit(self, func = None, p0 = None, use_weights = False):
        if not func:
            func = self.fitted_function
        if not use_weights:
            sigma = [1]*len(self.h)
        else:
            sigma = [1/w for w in self.calc_weights()]
        res, _ = curve_fit(
                f = func,
                xdata = self.h, 
                ydata = self.phi, 
                p0 = p0,
                sigma = sigma
                )
        return list(res)

    def calc_weights(self):
        return [(1/h_i)/sum(1/h_j for h_j in self.h) for h_i in self.h]

    def check_roache(self):
        msg = ''
        if not isclose(
                self.function_parameters[0], 
                self.phi_extrapolated
                ):
            msg = '. '.join(['Non-monotonic convergence',
                'Roache\'s method and fitted curve do not correspond. '
                ])
        apparent_order = self.calc_apparent_order()
        if apparent_order < 0.5 or apparent_order > 2:
            msg = ''.join([msg, 
                'Observed convergence is outside recommended bounds. ',
                'Consider adding solutions from additional grids.'
                ])
        if msg:
            warn(msg)
      
    def use_eca_hoekstra(self):
        safety_factor = 1.25
        functions = [self.fitfunc_unknown_order]
        fits = [f for i, f in enumerate(
            self.fit_weighted_unweighted(functions[0])
            ) if f and f[2] >= 0.5]
        if not True in [f[2] <= 2.0 for f in fits]:
            # If none of the fits are within required bounds
            functions = [self.fitfunc_1st_order, self.fitfunc_2nd_order]
            if not True in [f[2] <= 2.1 for f in fits]:
                # Additional tolerance for change in the safety factor
                safety_factor = 3
            if not True in [f[2] > 2 for f in fits]:
                # If none observed orders of convergence are larger than 2,
                # add mixed fit
                functions.append(self.fitfunc_1st_n_2nd_order)
            fits = [] # Discard fits of wrong order
            for func in functions:
                for f in self.fit_weighted_unweighted(func):
                    fits.append(f)
        min_stdev = min(f[-1] for f in fits if f)
        i, res = next((i, f) for i, f in enumerate(fits) 
                if f and f[-1] == min_stdev
                )
        self.phi_extrapolated = res[0]
        self.fitted_function = functions[int(i/2)]
        self.function_parameters = res[:-1]
        self.uncertainties = self.calc_uncertainties(res[-1], safety_factor)

    def fit_weighted_unweighted(self, func = None):
        # Returns regression parameters and standard deviation of fit as tuples
        fits = []
        if not func:
            func = self.fitted_function
        for b in [True, False]:
            try:
                fit = self.calc_curve_fit(func = func, use_weights = b)
                fit.append(self.calc_fit_stdev(
                    func(np.array(self.h), *fit), use_weights = b
                    ))
                fits.append(fit)
            except (RuntimeError, RuntimeWarning, OptimizeWarning):
                fits.append([])
        return fits

    def calc_fit_stdev(self, yfit, order = 0, use_weights = False):
        if order in (1, 2):
            len_correct = 2
        else:
            len_correct = 3
        if use_weights:
            nw = len(self.h)*np.array(self.calc_weights())
        else:
            nw = 1
        res = np.sqrt(
                np.sum(nw*(np.array(self.phi) - np.array(yfit))**2)
                / (len(self.h) - len_correct)
                )
        return float(res)

    def fitfunc_unknown_order(self, h, phi_0, alpha, p):
        return phi_0 + alpha*h**p

    def fitfunc_1st_order(self, h, phi_0, alpha):
        return phi_0 + alpha*h

    def fitfunc_2nd_order(self, h, phi_0, alpha):
        return phi_0 + alpha*h**2

    def fitfunc_1st_n_2nd_order(self, h, phi_0, alpha1, alpha2):
        return phi_0 + alpha1*h + alpha2*h**2

    def calc_uncertainties(self, stdev, safety_factor):
        d_phi = self.calc_data_range_parameter()
        phi_fit = [self.fitted_function(h_i, *self.function_parameters)
                for h_i in self.h
                ]
        epsilons = [abs(phi_fit_i - self.phi_extrapolated)
                for phi_fit_i in phi_fit
                ]
        common = [stdev + abs(phi_i - phi_fit_i) 
                for phi_i, phi_fit_i in zip(self.phi, phi_fit)
                ]
        if stdev < d_phi:
            return [safety_factor*e + c for e, c in zip(epsilons, common)]
        else:
            return [3*stdev/d_phi*(e + c) for e, c in zip(epsilons, common)]

    def calc_data_range_parameter(self):
        return (max(self.phi) - min(self.phi))/(len(self.h) - 1)

    def calc_extrapolated_errors(self):
        phi_ext = self.phi_extrapolated
        return [phi_i - phi_ext for phi_i in self.phi]
